from warnings import warn
import subprocess
from multiprocessing.connection import Client, Connection
from time import sleep
import random

import sqlalchemy as sa
import pandas as pd
import numpy as np
import numpy.typing as npt
import scipy
import mikeio1d as mi


class DrainageSystemData:
    """Object for static drainage system data. Takes care of reading and holding data. Also some calculations."""

    # Volume is checked to be <=VOLUME_CHECK_TOLERANCE_FACTOR*full_volume,
    # to make sure that the values are not completely wrong. The tolerance is used because it may be slightly off.
    VOLUME_CHECK_TOLERANCE_FACTOR = 1.1

    def __init__(self, model_sql_path: str, r1d_path: str) -> None:
        """Load data.
        Args:
            model_sql_path (str): path to the sql of the mike model.
            r1d_path (str): path to a result file.
        """
        # model data is saved as a sqlite, which we can read from
        self.model_db = sa.create_engine("sqlite:///" + model_sql_path)
        # Nodes and reaches are read from the res1d file.
        # A reach in this file can be any connection between nodes, including weirs, pumps, etc. Therefore this is simpler than the model sqlite.
        # for reference on reading the data see:
        # https://github.com/DHI/mikeio1d/discussions/53#discussion-5231633
        self.nwr = mi.Res1D(r1d_path)
        self.xs_reader = _connect_xs_reader(model_sql_path)
        self.catchment_df = _read_catch(self.model_db)
        self.node_df = _read_nodes(self.nwr, self.model_db)
        # this is later needed to get the start and end nodes of reaches, where the r1d only contains an index as a number
        self._r1d_node = self.node_df.reset_index()[["r1d_index", "name"]].set_index(
            "r1d_index"
        )["name"]
        self.manhole_df = _read_manhole(self.model_db)
        self.basin_df, self.basin_geometry = _read_basin(self.model_db)
        self.node_df["full_volume"] = 0.0
        self.node_df["full_volume"].update(self.manhole_df["full_volume"])
        self.node_df["full_volume"].update(self.basin_df["full_volume"])
        self.reach_df = _read_reach(self.nwr, self._r1d_node, self.xs_reader)
        self.weir_df = _read_weir(self.model_db)

    def calc_node_volume(self, node:str, water_level:float):
        """calculate volume for a node at specified water level

        Args:
            node (str): name of the node
            water_level (float): water_level (m), zero point is not the node bottom, but a general one

        Returns:
            float: volume (m³)
        """
        node_type = self.node_df.loc[node, "type"]
        if node_type == "Manhole":
            effective_water_level = np.clip(
                water_level,
                self.manhole_df.loc[node, "bottom_level"],
                self.manhole_df.loc[node, "ground_level"],
            )
            volume = (
                (effective_water_level - self.manhole_df.loc[node, "bottom_level"])
                * (self.manhole_df.loc[node, "diameter"] / 2) ** 2
                * np.pi
            )
            assert (
                volume
                <= self.VOLUME_CHECK_TOLERANCE_FACTOR
                * self.manhole_df.loc[node, "full_volume"]
            ), f"Volume for node {node} too large. Got {volume} expected<={self.manhole_df.loc[node,'full_volume']}"
        elif node_type == "Basin":
            bg = self.basin_geometry.query("node_name==@node")
            volume = (bg["h"].clip(upper=water_level).diff() * bg["surface_area"]).sum()
            assert (
                volume
                <= self.VOLUME_CHECK_TOLERANCE_FACTOR
                * self.basin_df.loc[node, "full_volume"]
            ), f"Volume for node {node} too large. Got {volume} expected<={self.basin_df.loc[node,'full_volume']}"
        else:
            volume = 0
        return volume

    def calc_reach_volume(self, reach, water_level):
        """calculate volume for a reach at specified water level. Assume that the water is flat.

        Args:
            reach (str): name of the reach
            water_level (float): water_level (m), zero point is not the reah bottom, but a general one

        Returns:
            float: volume (m³)
        """
        # get volume by numericaly integration the cross sections over the reach length
        STEP_SIZE = 1.0
        reach_length = self.reach_df.loc[reach, "length"]
        chainages_desired = np.append(
            np.arange(0, reach_length, STEP_SIZE), [reach_length]
        )
        self.xs_reader.send(("xs_areas_at", reach, water_level, chainages_desired))
        rsp = self.xs_reader.recv()
        if isinstance(rsp, Exception):
            warn(f"Could not read volume for {reach}. {rsp}")
            return 0
        else:
            chainages, areas = np.array(rsp)
            area_finite = np.isfinite(areas)
            volume = scipy.integrate.trapezoid(
                y=areas[area_finite], x=chainages[area_finite]
            )
            assert (
                volume
                <= self.VOLUME_CHECK_TOLERANCE_FACTOR
                * self.reach_df.loc[reach, "full_volume"]
            ), f"Volume for reach {reach} too large. Got {volume} expected<={self.reach_df.loc[reach,'full_volume']}"
            return volume

    def xs_area_at(
        self, reach: str, waterlevel: float, chainages: list[float]
    ) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
        self.xs_reader.send(('xs_areas_at', reach, waterlevel, chainages))
        rsp = self.xs_reader.recv()
        if isinstance(rsp,Exception):
            raise rsp
        else:
            return rsp


def _connect_xs_reader(model_sql_path: str):
    """Conect to xs_reader"""
    # For data on pipe cross sections, we need to use the mike1d api.
    # This does not seem to work when also importing mike1d.
    # Therefore use a multiprocessing.Listener, which runs in its own process with its own imports, and reads the file.

    # address for connection to reader. Getting a free port is somewhat complicated, but just using a random one should usually work.
    reader_address = ("localhost", random.randint(6000, 6100))
    auth_key = b"reach_reader"
    # Start the listener in its own console window.
    # normal
    # subprocess.Popen(
    #     [
    #         "python",
    #         "-m",
    #         "entfrachten.analysis.cross_section_info_loader",
    #         model_sql_path,
    #         "-a",
    #         str(reader_address),
    #     ],
    #     creationflags=subprocess.CREATE_NEW_CONSOLE,
    # )
    # for keeping cmd open if an error occurs use this, only for windows
    subprocess.Popen(
        f'cmd /k "python  -m entfrachten.analysis.cross_section_info_loader "{model_sql_path}" -a "{reader_address}""', creationflags=subprocess.CREATE_NEW_CONSOLE
    )
    # Listener may take some time to start, try to connect until it is accepted
    for n_try in range(20):
        try:
            # try to connect to server
            xs_reader = Client(reader_address, authkey=auth_key)
            break
        except (ConnectionRefusedError, FileNotFoundError) as e:
            # give server some time to start
            sleep(3)
    else:
        raise Exception(
            "Could not connect to listener. Check printed messages and that connection detail match up with cross_section_info_loader.py"
        )
    return xs_reader


def _read_catch(model_db: sa.Engine):
    # get catchment data from sqlite
    # msm_catchment contains some basic data. msm_catchcon defines where catchments are connected to.
    # note that area unit is m²
    with model_db.begin() as con:
        catchment_df = pd.read_sql(
            """
            SELECT catchid as id, nodeid as node, linkid as reach, area, ModelAImpArea as impervious_ratio , persons 
            FROM msm_catchcon LEFT JOIN msm_catchment ON msm_catchcon.catchid=msm_catchment.muid
            WHERE enabled=1""",
            con,
        )
        catchment_df["impervious_area"] = (
            catchment_df["area"] * catchment_df["impervious_ratio"]
        )
    return catchment_df


def _read_nodes(nwr: mi.Res1D, model_db: sa.Engine) -> pd.DataFrame:
    # for accessing data see: https://github.com/DHI/mikeio1d/discussions/53
    node_records = []
    for i, node in enumerate(list(nwr.data.Nodes)):
        node_records.append(
            {
                "name": node.ID,
                "r1d_index": i,
                "type": str(node.GetType()).split(".")[-1].removeprefix("Res1D"),
                "x": node.XCoordinate,
                "y": node.YCoordinate,
            }
        )
    node_df = pd.DataFrame.from_records(node_records)
    node_df = node_df.set_index("name")
    # some data is not in the res1d, use sqlite
    with model_db.begin() as con:
        node_sqlite = pd.read_sql(
            """
            SELECT muid as name, groundlevel as ground_level, invertlevel as bottom_level
            FROM msm_Node""",
            con,
        ).set_index("name")
    node_df = node_df.join(node_sqlite)
    return node_df


def _read_manhole(model_db: sa.Engine) -> pd.DataFrame:
    # read data in manholes. Those are denoted by typeno 1.
    with model_db.begin() as con:
        manhole_df = pd.read_sql(
            """
            SELECT muid as node_name, groundlevel as ground_level, invertlevel as bottom_level, diameter
            FROM msm_Node
            WHERE typeno=1""",
            con,
        ).set_index("node_name")
        manhole_df["full_volume"] = (
            (manhole_df["ground_level"] - manhole_df["bottom_level"])
            * (manhole_df["diameter"] / 2) ** 2
            * np.pi
        )
    return manhole_df


def _read_basin(model_db: sa.Engine) -> tuple[pd.DataFrame, pd.DataFrame]:
    # read basin data
    # typeno=2 denotes basins. The geometry is saved in an extra table
    with model_db.begin() as con:
        basin_df = pd.read_sql(
            """
            SELECT msm_node.muid as node_name, groundlevel as ground_level, invertlevel as bottom_level
            FROM msm_node
            WHERE msm_node.typeno=2""",
            con,
        ).set_index("node_name")
        # the geometry is saved in an extra table. This table is very generic, therefore columns are called value1, value2 etc.
        # The important columns for us are value1, which contains the height, and value3, which contains the surface area at the height of the same row.
        # Which column is which can be seen from the tooltips in Mike+.
        basin_geometry = pd.read_sql(
            """
            SELECT msm_node.muid as node_name, msm_node.geometryid as geometry_id, ms_tabd.value1 as h, ms_tabd.value3 as surface_area
            FROM msm_node LEFT JOIN ms_tabd ON msm_node.geometryid=ms_tabd.tabid
            WHERE msm_node.typeno=2
            ORDER BY node_name,h""",
            con,
        ).set_index("node_name")
    basin_df["full_volume"] = (
        (
            basin_geometry["h"].groupby(level="node_name").diff()
            * basin_geometry["surface_area"]
        )
        .groupby(level="node_name")
        .sum()
    )
    basin_df["geometry_id"] = (
        basin_geometry["geometry_id"].groupby(level="node_name").first()
    )
    return basin_df, basin_geometry


def _read_reach(
    nwr: mi.Res1D, r1d_node: pd.Series, xs_reader: Connection
) -> pd.DataFrame:
    reach_records = []
    for reach in list(nwr.data.Reaches):
        reach_records.append(
            {
                "name": reach.Name,
                "start": r1d_node[reach.StartNodeIndex],
                "end": r1d_node[reach.EndNodeIndex],
                "length": reach.Length,
            }
        )
    reach_df = pd.DataFrame.from_records(reach_records).set_index("name")
    # add type from name, everything  other than link has a prefix "{type}:"
    has_colon = reach_df.index.to_series().str.contains(":")
    reach_df.loc[~has_colon, "type"] = "Link"
    reach_df.loc[has_colon, "type"] = reach_df.index.to_series().str.split(
        ":", expand=True
    )[0]
    assert reach_df.index.is_unique
    # add cross_section area of reach
    for reach in reach_df.index:
        xs_reader.send(("full_xs_area", reach))
        a = xs_reader.recv()
        if not isinstance(a, float):
            a = 0
        reach_df.loc[reach, "xs_area"] = a
        reach_df.loc[reach, "full_volume"] = reach_df.loc[reach, "length"] * a
    return reach_df


def _read_weir(model_db: sa.Engine) -> pd.DataFrame:
    with model_db.begin() as con:
        weir_df = pd.read_sql(
            """
            SELECT muid as id, 'Weir:'||muid as reach, fromnodeid as start, tonodeid as end, crestlevel as crest_level
            FROM msm_weir;""",
            con,
        ).set_index("id")
    return weir_df
