"""Module with helper methods for dynamic analysis, i.e data that changes with time, like flow and concentrations. """
import typing

import pandas as pd
import numpy as np
import scipy.signal

from entfrachten.analysis import unit_factors as uf

_REDUCTION_FACTOR = 0.75
ids = pd.IndexSlice
# Parts that are considered for how water volume and tss mass are distributed.
dist_agg_parts = [
    "inflow",
    "storage",
    "overflow",
    "outflow",
]

# size of slices for volume sections
slice_mm = 1

def _calc_timestep_seconds(s:pd.Series):
    """Return series with timestep in seconds. Assume that the index levels are [place, event, datetime]."""
    return (
        s.index.to_frame()["datetime"]
        .groupby(["place", "event"])
        .diff()
        .bfill()
        .dt.total_seconds()
    )


def calc_raindepth(volume: pd.Series, sealed_area: float) -> pd.Series:
    """Convert volume to corresponding accumulated rain depth"""
    # Calculate the rainfall depth
    # m³/m2=m, multiply with 1000 to get mm
    return volume / sealed_area / _REDUCTION_FACTOR * 1000


def calc_volume_from_raindepth(rain_depth: pd.Series, sealed_area: float) -> pd.Series:
    """Convert volume to corresponding accumulated rain depth"""
    # Calculate the rainfall depth
    # m³/m2=m, multiply with 1000 to get mm
    return rain_depth * sealed_area * _REDUCTION_FACTOR / 1000


def volume_sections_from_raindepth(
    volume: pd.Series, sealed_area: float
) -> pd.Series:
    """Make sections based on volume converted to rain depth and returns the upper bound for volume(m³) in the section for each entry in volume.
    This is intended to make comparable volume sections for all storages."""
    rain_depth = calc_raindepth(volume, sealed_area)
    rain_slice = np.floor(rain_depth / slice_mm)
    rain_slice_end = rain_slice * slice_mm + slice_mm
    volume_ubound = calc_volume_from_raindepth(rain_slice_end, sealed_area)
    return volume_ubound


def sections_from_raindepth(
    volume: pd.Series, sealed_area: float
) -> pd.Series:
    """Make sections based on volume converted to rain depth and returns the upper bound for rain depth (mm) in the section for each entry in volume.
    This is intended to make comparable volume sections for all storages."""
    rain_depth = calc_raindepth(volume, sealed_area)
    rain_slice = np.floor(rain_depth / slice_mm)
    rain_slice_end = rain_slice * slice_mm + slice_mm
    return rain_slice_end


def volume_dist(
    inflow: pd.Series,
    outflow: pd.Series,
    overflow: pd.Series,
    storage_volume: pd.Series,
    sub_storage_volume: pd.Series,
    sub_attribute_to: typing.Mapping,
):
    """Calculate accumulated volume for all parts. Substorages get extra columns in the output."""
    #   inflow ---> storage -----> outflow
    #              ==weir==
    #                 |
    #                 V
    #             overflow
    # timestep in seconds
    timestep_sec = _calc_timestep_seconds(inflow)
    # Ignore negative flow, we want to group by incoming volume. Water going backwards (i.e out) is ignored.
    in_volume = inflow.clip(lower=0).groupby(["place", "event"]).cumsum() * timestep_sec
    out_volume = outflow.groupby(["place", "event"]).cumsum() * timestep_sec
    overflow = overflow.reindex_like(inflow).fillna(0)
    over_volume = overflow.groupby(["place", "event"]).cumsum() * timestep_sec
    # dist storage, use attribute_to as place and put each actual place into a new column.
    # This is not efficient, but makes it simpler to arrange the fitting data points.
    sub_storage_volume = (
        sub_storage_volume.copy()
        .rename("volume")
        .reset_index()
        .rename(columns={"place": "storage_part"})
    )
    sub_storage_volume["place"] = sub_storage_volume["storage_part"].map(
        sub_attribute_to
    )
    sub_storage_volume = (
        sub_storage_volume.set_index(["storage_part", "place", "event", "datetime"])
        .unstack("storage_part")
        .droplevel(0, axis="columns")
    )
    assert (
        len(
            sub_storage_volume.columns.intersection(
                ["inflow", "storage", "overflow", "outflow"]
            )
        )
        == 0
    )
    volume_dist = pd.DataFrame(
        {
            "inflow": in_volume,
            "storage": storage_volume,
            "overflow": over_volume,
            "outflow": out_volume,
        }
        | {c: sub_storage_volume[c] for c in sub_storage_volume.columns},
        index=in_volume.index,
    )
    volume_dist.columns.name = "part"
    return volume_dist


def load_dist(
    inflow_transport: pd.Series,
    outflow_transport: pd.Series,
    overflow_transport: pd.Series,
    storage_volume: pd.Series,
    storage_conc: pd.Series,
    sub_storage_volume: pd.Series,
    sub_storage_conc: pd.Series,
    sub_attribute_to: typing.Mapping,
) -> pd.DataFrame:
    """Calculate accumulated load for all parts. Substorages get extra columns in the output."""
    timestep_sec = _calc_timestep_seconds(inflow_transport)
    # calculate these parts from transport accumulation
    in_load = inflow_transport.groupby(["place", "event"]).cumsum() * timestep_sec
    out_load = outflow_transport.groupby(["place", "event"]).cumsum() * timestep_sec
    overflow_transport = overflow_transport.reindex_like(outflow_transport).fillna(0)
    overflow_load = overflow_transport.groupby(["place", "event"]).cumsum() * timestep_sec

    # load in storages is calculated from concentration and volume
    storage_load = storage_volume * uf.MGPL_TO_KGPM3 * storage_conc
    # like for the unused storage analysis volume, we need to handle the fact that
    # some storages are made up of multiple parts, which are defined in storage_dist
    sub_storage = sub_storage_volume * uf.MGPL_TO_KGPM3 * sub_storage_conc
    # now merge l_s_sp according to dist_storage[attribute_to] into l_s
    # same procedure as for unused storages
    # reform attribute_to to have a values for each row of dist_storage_load
    long_attribute_to = sub_storage.index.get_level_values("place").map(sub_attribute_to)
    # 'as_index=False' is not allowed on series. It is also not needed here, since this operation seems to leave the index as is.
    for storage, dist_storage_load_part in sub_storage.groupby(
        long_attribute_to
    ):
        # index slice of volume to insert data into l_s
        to_slice = ids[storage, :, :]
        # sum loads up, per event and datetime pair, i.e sum over the places
        l_sum = dist_storage_load_part.groupby(["event", "datetime"]).sum()
        # add attribute_to as index level (by misusing concat) and reorder to match l_s
        l_sum = pd.concat([l_sum], keys=[storage], names=["place"])
        l_sum = l_sum.reorder_levels(storage_load.index.names)
        storage_load.loc[to_slice] = l_sum
    # now that l_s_sp was aggregated into l_s, change its form to also make the detailed values available, same as for volume
    sub_storage = (
        sub_storage.rename("load").reset_index().rename(columns={"place": "storage_part"})
    )
    sub_storage["place"] = sub_storage["storage_part"].map(sub_attribute_to)
    sub_storage = (
        sub_storage.set_index(["storage_part", "place", "event", "datetime"])
        .unstack("storage_part")
        .droplevel(0, axis="columns")
    )
    load_dist = pd.DataFrame(
        {
            "inflow": in_load,
            "storage": storage_load,
            "overflow": overflow_load,
            "outflow": out_load,
        }
        | {c: sub_storage[c] for c in sub_storage.columns},
        index=in_load.index,
    )
    load_dist.columns.name = "part"
    return load_dist

def volume_load_dist_delta(volume_dist,load_dist,sealed_area_m2):
    """Calculates changes over volume sections from accumulated load and volume."""
    v_in = volume_dist["inflow"]
    # note that this function assigns a section to every value in the input
    volume_sections = pd.Series(index=v_in.index, dtype=v_in.dtype)
    section_sizes = pd.Series(index=sealed_area_m2.index, dtype=float, data=None)
    # the area in m2 is later needed to convert volume to a corresponding rain depth
    
    for p in volume_sections.index.get_level_values("place").unique():
        # using p instead of [p] would make the assignment fail, probably because the version with [p] removes p from the index
        sections_p = volume_sections_from_raindepth(v_in.loc[[p], :, :], sealed_area_m2[p])
        volume_sections.loc[[p], :, :] = sections_p
        section_sizes[p] = sections_p.iloc[0]
    assert (volume_sections > 0).all()
    # check where the volume sections are not monotonic increasing, i.e incoming volume decreases enough to be in a smaller section
    is_increasing = volume_sections.groupby(["place","event"]).is_monotonic_increasing
    if ~is_increasing.sum()>0:
        print("volume sections not monotonic increasing:")
        print((~is_increasing).groupby("place").sum())
    # get indices where v_in is nearest to the section value per place, event and section.
    section_delta = (volume_sections - v_in).abs()
    section_idx = (
        pd.DataFrame(
            {
                "delta": section_delta,
                "volume_section": volume_sections,
            }
        )
        .groupby(["place", "event", "volume_section"])
        .idxmin()["delta"]
    )
    # first get the accumulated values
    # some values (like storage) may no start at 0, correct that
    load_dist_acc = load_dist-load_dist.groupby(["place","event"]).transform("first")
    volume_dist_acc = volume_dist-volume_dist.groupby(["place","event"]).transform("first")
    load_dist_acc = load_dist_acc.loc[section_idx, :]
    volume_dist_acc = volume_dist_acc.loc[section_idx, :]
    # section_idx values are where the values are found, the index contains the volume sections
    load_dist_acc.index = section_idx.index
    volume_dist_acc.index = section_idx.index
    # now convert to difference
    # diff is nan for first value. here we insert the original value, since we can assume 0 as first value
    load_dist_d = load_dist_acc.groupby(["place", "event"]).diff().fillna(load_dist_acc)
    volume_dist_d = (
        volume_dist_acc.groupby(["place", "event"]).diff().fillna(volume_dist_acc)
    )
    # compare intended section size to the actual changes in volume
    # this is expected to not always match because the timesteps may not match up with desired volume changes.
    section_sizes_each_section = volume_dist_d.index.get_level_values("place").map(section_sizes)
    in_section_ratio = volume_dist_d["inflow"]/section_sizes_each_section
    print("actual inflow change/desired section size")
    print(in_section_ratio.describe())
    # correct values
    volume_dist_d=volume_dist_d.div(in_section_ratio,axis="index")
    load_dist_d=load_dist_d.div(in_section_ratio,axis="index")
    return volume_dist_d, load_dist_d

def consecutive_true_size(bool_series: pd.Series):
    """Calculate the maximum size of consecutive true values in bool_series.
    Size is calculated from the differences of index-values, i.e for a single true value it will be zero.
    For int index size is int, for Timestamp size is timedelta, etc.
    Can be used for calculating event or gap durations.
    """
    prev_true = bool_series.shift(1, fill_value=False)
    next_true = bool_series.shift(-1, fill_value=False)
    first_true = bool_series & (~prev_true)
    last_true = bool_series & (~next_true)

    times = bool_series.index.get_level_values("datetime")
    first_true_pos = pd.Series(
        data=pd.NaT, index=bool_series.index,
    )
    last_true_pos = pd.Series(
        data=pd.NaT, index=bool_series.index,
    )

    first_true_pos[first_true] = times[first_true]
    last_true_pos[last_true] = times[last_true]

    first_true_pos = first_true_pos.ffill()
    last_true_pos = last_true_pos.bfill()

    true_size = last_true_pos - first_true_pos
    # set size to nan where bool_series is false
    true_size[~bool_series] = None
    return true_size

def calc_lag(a:pd.Series,b:pd.Series,window_steps:int,vmin):
    """Calculate the lag from window wise cross correlation.
    vmin is used to ignore parts where a and b are lower than vmin, there nan is returned
    Returns: The lag in number of timesteps"""
    assert a.index.equals(b.index)
    a_numpy = a.to_numpy().astype("double")
    b_numpy = b.to_numpy().astype("double")
    # set the maximum amount of lag in number of timesteps
    # set size of the window as distance from the middle to the edges.
    win_pad = window_steps
    # save lag for each point in a series.
    lag = pd.Series(
        data=np.nan,
        index=a.index,
        name="lag",
        dtype=float,
    )
    # calculate lag for each position, if one of the values is larger than vmin
    for t in range(len(lag)):
        if (a_numpy[t] >= vmin) or (b_numpy[t] >= vmin):
            # Start and end of windwow, make sure indices are inside the valid range.
            win_start = max(0, t - win_pad)
            win_end = min(t + win_pad, len(a_numpy) - 1)
            # extract window data
            res_part = a_numpy[win_start : win_end + 1]
            ref_part = b_numpy[win_start : win_end + 1]
            # use scyipy.signal to calculate correlation
            correlation_mode: str = "same"
            correlation = scipy.signal.correlate(
                res_part, ref_part, mode=correlation_mode
            )
            # signal.correlation_lags tells  us, how the index of the correlation result matches a lag in the original data.
            # For our cases the lags shoudl just be and array of the from [-n,...,0,...,n], but this method handles edge cases
            # and different correlation_modes for us.
            all_lags = scipy.signal.correlation_lags(
                res_part.size, ref_part.size, mode=correlation_mode
            )
            # use is_lag_valid to only search lags smaller than the limit for the best result.
            is_lag_valid = np.abs(all_lags) <= window_steps
            idx_lag = all_lags[is_lag_valid][np.argmax(correlation[is_lag_valid])]
            lag.iloc[t] = idx_lag
    return lag