"""Module to make folium maps of the drainage system."""
# TODO less copy paste
import collections
from warnings import warn
import numpy as np
import pandas as pd
import geopandas as gpd
import shapely
import shapely.affinity as saf
import folium
import branca
import matplotlib as mpl


from tqdm import tqdm

from entfrachten.analysis.drainage_sysytem_data import DrainageSystemData
from entfrachten.analysis.drainage_system_graph import DrainageSystemGraph


# colormap for volume
_invalid_value_color = "#999999"
_volume_over_color = "#000000"
_volume_cmap = mpl.colormaps.get_cmap("plasma_r").copy()
_volume_cmap.set_bad(_invalid_value_color)
_volume_cmap.set_under(_invalid_value_color)
_volume_cmap.set_over(_volume_over_color)



def _map_df(sd:DrainageSystemData) -> gpd.GeoDataFrame:
    """Prepares a GeoDataframe for use as a map."""
    # to_geopandas translates the network to a geopandas dataframe.
    # This is very handy for visualization.
    map_df = sd.nwr.network.to_geopandas()
    map_df = map_df.merge(sd.node_df[["type"]],left_on="name",right_on="name",how="left")
    is_reach = map_df["group"]=="Reach"
    has_colon = map_df["name"].str.contains(":")
    map_df.loc[is_reach&~has_colon,"type"]="Link"
    map_df.loc[is_reach&has_colon,"type"]=map_df["name"].str.split(":",expand=True)[0]
    return map_df # type: ignore

def type_map(sd:DrainageSystemData,storage_df:pd.DataFrame):
    """Draw a map of the network colored by the type of the element."""
    color_map = {
        "Manhole":"#0000ff",
        "Basin":"#00ff00",
        "SewerJunction":"#888888",
        "Outlet":"#ff0000",
        "Link":"#0000ff",
        "Weir":"#ff0000",
        "Valve":"#8888ff",
        "Pump":"#ffff00",
        "Orifice":"#000000",
    }
    map_df=_map_df(sd)
    map_df["color"]=map_df["type"].map(color_map)
    # draw interesting parts larger
    interesting = ((map_df["type"]!="Link")&(map_df["type"]!="Manhole")&(map_df["type"]!="SewerJunction"))|map_df["name"].isin(storage_df["node"])
    m = map_df.loc[~interesting,:].explore(color=map_df.loc[~interesting,"color"],marker_kwds={"radius":1.5})
    map_df.loc[interesting,:].explore(color=map_df.loc[interesting,"color"],marker_kwds={"radius":6},m=m)
    _add_storage_markers(m, map_df[map_df["group"] == "Node"], storage_df)
    return m


def _add_storage_markers(m,node_gdf,storage_df):
    """Add a marker for storage nodes."""
    node_gdf=node_gdf.reset_index()
    node_icon = folium.Icon(icon_color="green", icon="tint")
    storage_map_df = node_gdf[
        node_gdf["name"].isin(storage_df["node"])
    ]
    for s, n in storage_df["node"].items():
        storage_map_df.loc[node_gdf["name"] == n, "storage"] = s
    storage_map_df.explore(
        m=m, marker_type="marker", marker_kwds={"icon": node_icon, "radius": 4}
    )

def partition_map(sd:DrainageSystemData,storage_df:pd.DataFrame,graph:DrainageSystemGraph):
    """Draw the partition of network that is used to sum up data."""
    neutral_color = "#AAAAAA"
    mix_color = "#000000"
    map_df = _map_df(sd)
    map_df["color"] = neutral_color
    color_cycle = [
        "#FBB3C1",
        "#E7C095",
        "#B9CF8E",
        "#81D7B4",
        "#79D4E1",
        "#B6C5F8",
        "#ECB5EB",
        "#A86371",
        "#95703F",
        "#6A7F36",
        "#158764",
        "#008591",
        "#6576A8",
        "#9C639B",
    ]
    map_df["storages"] = "{"
    node_storage = storage_df.reset_index()[["node","name"]].set_index("node")["name"]
    if len(storage_df)<=len(color_cycle):
        warn("Not enough colors to show all storages!")
    for n, c in tqdm(list(zip(storage_df["node"], color_cycle))):
        s = node_storage[n]
        next_up_nodes, next_up_reaches = graph.find_upstream_nodes_reaches(
            n, storage_df["node"]
        )
        n_match = (map_df["name"].isin(next_up_nodes) & (map_df["group"] == "Node")) | (
            map_df["name"].isin(next_up_reaches) & (map_df["group"] == "Reach")
        )
        map_df.loc[n_match, "storages"] += "," + s
        is_colored = map_df["color"] != neutral_color
        map_df.loc[n_match, "color"] = c
        map_df.loc[n_match & is_colored, "color"] = mix_color
    map_df["storages"] += "}"
    is_storage = map_df["name"].isin(storage_df["node"]) & (map_df["group"] == "Node")
    map_df_fat = map_df.loc[is_storage]
    map_df_thin = map_df.loc[~is_storage]
    m = map_df_thin.explore(
        color=map_df_thin["color"], marker_kwds={"radius": 1.5}, tiles="Cartodb Positron"
    )
    map_df_fat.explore(color=map_df_fat["color"], marker_kwds={"radius": 6}, m=m)
    _add_storage_markers(m,map_df,storage_df)
    return m


def _vol_to_col(v, vmax):
    """Map volume to color."""
    assert np.ndim(v) == 0, f"{v} not scalar"
    if v <= 0:
        return _invalid_value_color
    rgba = _volume_cmap(v / vmax)
    return mpl.colors.rgb2hex(rgba)


def volume_map(sd:DrainageSystemData,storage_df):
    """Make a map that shows reach cross section and node volume."""
    # convert values to this crs so that coordinates are in meters
    metric_crs = "EPSG:25834"
    reach_max_width = 10
    node_max_radius = 20
    normalization_quantile = 0.99
    max_xs = sd.reach_df["xs_area"].quantile(normalization_quantile)
    node_gdf = sd.nwr.nodes.to_geopandas().to_crs("EPSG:25834").set_index("name")
    node_gdf = node_gdf.drop(columns="group").join(
        sd.node_df.drop(columns=["x", "y", "r1d_index"])
    )
    # make a version of node_gdf where the area is proportional to the volume
    node_vol_gdf = node_gdf.copy()
    node_vol_gdf["full_volume"] = 0.0
    min_radius = 0.5
    node_vol_gdf["map_radius"] = min_radius
    for n in node_vol_gdf.index:
        node_vol_gdf.loc[n, "full_volume"] = sd.calc_node_volume(
            n, node_vol_gdf.loc[n, "ground_level"]
        )
    max_node_vol = node_vol_gdf["full_volume"].quantile(normalization_quantile)
    for n in node_vol_gdf.index:
        # a==v <=> r**2==v <=> r == v**0.5
        t = node_vol_gdf.loc[n, "full_volume"]
        node_vol_gdf.loc[n, "map_radius"] = max(
            (node_max_radius * t / max_node_vol) ** 0.5, min_radius
        )

    node_vol_gdf.geometry = node_vol_gdf.geometry.buffer(
        node_vol_gdf["map_radius"], resolution=8
    )
    reach_vol_gdf = gpd.GeoDataFrame(
        data=sd.reach_df, geometry=[None] * len(sd.reach_df), crs=metric_crs
    )
    for r in reach_vol_gdf.index:
        # make a triangle where one point is at end and one side at start
        # area should be proportional to volume
        # a~v <=> w*l ~ xs*l <=>  w~xs
        s = node_gdf.geometry.loc[reach_vol_gdf.loc[r, "start"]]
        e = node_gdf.geometry.loc[reach_vol_gdf.loc[r, "end"]]
        width = reach_max_width * sd.reach_df.loc[r, "xs_area"] / max_xs
        scale = width / shapely.distance(s, e)
        e_scaled = saf.scale(e, xfact=scale, yfact=scale, origin=s)
        s1 = saf.rotate(e_scaled, 90, origin=s)
        s2 = saf.rotate(e_scaled, -90, origin=s)
        triangle = shapely.Polygon([s1, s2, e])
        reach_vol_gdf.loc[r, "geometry"] = triangle
    # add colors
    max_col_vol = pd.concat(
        [node_vol_gdf["full_volume"], reach_vol_gdf["full_volume"]]
    ).quantile(normalization_quantile)
    node_vol_gdf["color"] = node_vol_gdf["full_volume"].apply(_vol_to_col, vmax=max_col_vol)
    reach_vol_gdf["color"] = reach_vol_gdf["full_volume"].apply(
        _vol_to_col, vmax=max_col_vol
    )
    cm_pad = 0.1
    cm_samples = np.linspace(-cm_pad * max_col_vol, (1 + cm_pad) * max_col_vol, 400)
    map_colormap = branca.colormap.LinearColormap(
        [_vol_to_col(v, vmax=max_col_vol) for v in cm_samples],
        vmin=cm_samples[0],
        vmax=cm_samples[-1],
        caption="Volume(m³)\n Area is proportional to volume, but not comparable for reaches and nodes",
    )
    volume_map = node_vol_gdf.explore(
        color=node_vol_gdf["color"], tiles="Cartodb Positron", popup=True, max_zoom=24
    )
    reach_vol_gdf.explore(color=reach_vol_gdf["color"], m=volume_map, popup=True)
    _add_storage_markers(volume_map, node_gdf, storage_df)
    map_colormap.add_to(volume_map)
    return  volume_map


def _time_to_col(t, vmax):
    """Map time to color"""
    time_cm = mpl.colormaps.get_cmap("viridis")
    if vmax == 0 or vmax == pd.Timedelta("0s"):
        return "#999999"
    assert np.ndim(t) == 0, f"{t} not scalar"
    rgba = time_cm(t / vmax)
    return mpl.colors.rgb2hex(rgba)


def make_path_time_map(sd:DrainageSystemData, path, max_time=None):
    """Draw flowtime on map, normalized with max from path. Reaches in Path are drawn thicker."""
    map_df_diff = _map_df(sd).set_index("name")
    map_df_diff["length"] = sd.reach_df["length"]
    map_df_diff["max_velocity"] = sd.reach_df["max_velocity"]
    map_df_diff["flow_time(min)"] = sd.reach_df["flow_time"] / pd.Timedelta(minutes=1)
    map_df_diff["flow_time_norm"] = (
        sd.reach_df["flow_time"] / pd.Timedelta(seconds=1)
    ) / sd.reach_df["length"]
    if max_time == None:
        max_time = map_df_diff.loc[path, "flow_time_norm"].max()
    map_df_diff["color"] = map_df_diff["flow_time_norm"].apply(
        _time_to_col, vmax=max_time
    )
    map_df_path = map_df_diff.loc[path].reset_index()
    map_df_rest = map_df_diff.loc[map_df_diff.index.difference(path)].reset_index()
    m = map_df_rest.explore(
        color=map_df_rest["color"],
        tiles="Cartodb Positron",
        popup=True,
        max_zoom=24,
        style_kwds=dict(weight=2, opacity=0.7, fillOpacity=0.6),
        marker_kwds=dict(radius=1, stroke=False, fill_opacity=0.2),
    )
    map_df_path.explore(
        color=map_df_path["color"], popup=True, style_kwds=dict(weight=5), m=m
    )
    # add a colormap legend
    cm_samples = np.linspace(0, max_time, 100)
    map_colormap = branca.colormap.LinearColormap(
        [_time_to_col(t, vmax=max_time) for t in cm_samples],
        vmin=cm_samples[0],
        vmax=cm_samples[-1],
        caption="normalized flow time (s/m)",
    )
    map_colormap.add_to(m)
    return m
